﻿#include <windows.h>
#include <iterator>
#include <utility>
#include <memory>
#include <string>
#include <vector>
#include <array>
#include <unordered_set>
#include <unordered_map>
#include <algorithm>

#include "MQPlugin.h"

#ifndef _CMQObjectDefine
#define _CMQObjectDefine

#define STEPIDX(index, step, size) ((index + step) % size)
#define NEXTIDX(index, size) (STEPIDX(index, 1, size))
#define NEXTIDX3(index) (STEPIDX(index, 1, 3))
#define PREVIDX(index, size) ((index + size - 1) % size)
#define PREVIDX3(index) (PREVIDX(index, 3))

namespace cmq
{
	using oindex_t = int;
	using vindex_t = int;
	using findex_t = int;
	using lindex_t = int64_t;
	using cindex_t = int;
	using mindex_t = int;
	using flindex_t = int64_t;

	class InvalidException
	{
	};

	class CMQObject;

	lindex_t makeLineId(vindex_t v1, vindex_t v2);
	lindex_t makeLineId(std::pair<vindex_t, vindex_t> line);
	std::pair<vindex_t, vindex_t> lineId2VertexPair(lindex_t lindex);

	flindex_t makeFaceLineId(findex_t f, int l);
	std::pair<findex_t, int> flId2FaceLinePair(flindex_t flindex);

	// 三角ポリゴンの頂点インデックス格納用
	struct TrianglePoints
	{
		TrianglePoints() = default;
		TrianglePoints(int v1, int v2, int v3) { idx = { v1, v2, v3 }; }
		std::array<int, 3> idx = { -1, -1, -1 };
	};
	
	// 辺指定インデックスのため、二つのintを持つインデックス
	class CMQpair
	{
	public:
		bool operator==(const CMQpair &other) const { return v == other.v; }
		bool operator!=(const CMQpair &other) const { return !operator==(other); }

		std::size_t hash() const {
			return std::hash<int>()(v[0]) ^ (std::hash<int>()(v[1]));
		}
		struct HashFunctor {
			std::size_t operator()(const CMQpair &k) const {
				return k.hash();
			}
		};

	protected:
		CMQpair() = default;
		CMQpair(int v1, int v2) : v(std::array<int, 2>{v1, v2}) { }
		CMQpair(const CMQpair &other) : v(other.v) { }

		int v1_() const { return v[0]; }
		int v2_() const { return v[1]; }
		void v1_(int v1_) { v[0] = v1_; }
		void v2_(int v2_) { v[1] = v2_; }

	private:
		std::array<int, 2> v = { -1, -1 };
	};

	// 2頂点から辺を指定するインデックス
	class CMQLindex : public CMQpair
	{
	public:
		CMQLindex() = default;
		CMQLindex(vindex_t v1, vindex_t v2) : CMQpair(min(v1, v2), max(v1, v2)) { }
		CMQLindex(const CMQLindex &other) : CMQpair(other) { }

		vindex_t v1() const { return CMQpair::v1_(); }
		vindex_t v2() const { return CMQpair::v2_(); }
		void v1(vindex_t v1_) { CMQpair::v1_(v1_); }
		void v2(vindex_t v2_) { CMQpair::v2_(v2_); }
	};

	// 面インデックスと辺番号から辺を指定するインデックス
	class CMQFLindex : public CMQpair
	{
	public:
		CMQFLindex() = default;
		CMQFLindex(vindex_t v, int l) : CMQpair(v, l) {}
		CMQFLindex(const CMQFLindex &other) : CMQpair(other) { }

		vindex_t v() const { return CMQpair::v1_(); }
		int l() const { return CMQpair::v2_(); }
		void v(vindex_t v_) { CMQpair::v1_(v_); }
		void l(int l_) { CMQpair::v2_(l_); }
	};

	class CMQPoint
	{
	public:
		static float innercross(MQPoint p1, MQPoint p2) {
			return p1.x * p2.x + p1.y * p2.y + p1.z * p2.z;
		}

		static MQPoint outercross(MQPoint p1, MQPoint p2) {
			return MQPoint(p1.y * p2.z - p1.z * p2.y, p1.z * p2.x - p1.x * p2.z, p1.x * p2.y - p1.y * p2.x);
		}
	};

	class CMQVertex
	{
	public:
		CMQVertex(const MQPoint &p, CMQObject *obj, vindex_t index) : m_p(p), m_obj(obj), m_index(index) {}
		~CMQVertex() {
			m_index = -1;
			m_obj = nullptr;
			m_p.zero();
		}
		const MQPoint &p() const { return m_p; }
		bool isSelect() const;
		void addSelect() const;
		void deleteSelect() const;
		std::vector<findex_t> getVertexRelatedFaces() const;
		int getVertexRefCount() const;
	private:
		vindex_t m_index;
		CMQObject *m_obj;
		MQPoint m_p;
	};

	class CMQFace
	{
	public:
		CMQFace(CMQObject *obj, findex_t index) : m_obj(obj), m_index(index) {}
		~CMQFace() {
			m_obj = nullptr;
			m_index = -1;
		}
		bool isSelect() const;
		void addSelect() const;
		void deleteSelect() const;
		int getFacePointCount() const;
		std::vector<vindex_t> getFacePointArray(std::vector<int> &vertexes = std::vector<int>(5)) const;
		void copyFaceAttribute(const CMQObject &object, findex_t findex2);
	private:
		findex_t m_index;
		CMQObject *m_obj;
	};

	class CMQCoordinate
	{
	public:
		CMQCoordinate() : index_(-1) { cood_.u = cood_.v = 0.0f; }
		CMQCoordinate(const CMQCoordinate &cood) : index_(cood.index_), cood_(cood.cood_) {}
		CMQCoordinate &operator=(const CMQCoordinate &cood) { index_ = cood.index_; cood_ = cood.cood_; return *this; }

		CMQCoordinate(cindex_t index, const MQCoordinate &cood) : index_(index), cood_(cood) {}
		bool operator==(const CMQCoordinate &other) const { return index_ == other.index_; }
		static bool uvcomp(const CMQCoordinate &left, const CMQCoordinate &right)
		{
			return pow2(left.cood_.u - right.cood_.u) + pow2(left.cood_.v - right.cood_.v) < pow2(uvdelta);
		}
		const float &u() const { return cood_.u; }
		const float &v() const { return cood_.v; }
		float &u() { return cood_.u; }
		float &v() { return cood_.v; }
		cindex_t getIndex() const { return index_; }
		const MQCoordinate &getCoordinate() const { return cood_; }
		MQCoordinate &getCoordinate() { return cood_; }
		void set(const MQCoordinate &coord) { cood_ = coord; }
	private:
		cindex_t index_;
		MQCoordinate cood_;

		static const float uvdelta; // UV値同一判定用
		static float pow2(float v) { return v * v; }
	};

	template <typename T, typename C>
	class CMQIterator : public std::iterator<std::bidirectional_iterator_tag, int>
	{
	public:
		friend class CMQObject;

		CMQIterator(const CMQIterator &iter) : m_container(iter.m_container), m_index(iter.m_index) {
		}

		CMQIterator &operator++() {
			m_index = next(m_index, true);
			return *this;
		}

		CMQIterator operator++(int) {
			CMQIterator origiter(*this);
			this->operator++();
			return origiter;
		}

		CMQIterator &operator--() {
			m_index = next(m_index, false);
			return *this;
		}

		CMQIterator operator--(int) {
			CMQIterator<T> origiter(*this);
			this->operator--();
			return origiter;
		}

		bool operator!=(const CMQIterator &iter) const { return m_index != iter.m_index; }
		bool operator==(const CMQIterator &iter) const { return !(this->operator!=(iter)); }

		//	const T &operator*() const {return get();}
		//	T &operator*() {return const_cast<T&>(get());}
		std::shared_ptr<T> operator->() const {
			std::shared_ptr<T> ptr(getp());
			return ptr;
		}
		//	T *operator->() {return &get(m_index);}

		int index() const { return m_index; }
		//T &get() const;
		std::shared_ptr<T> getp() const;

	private:
		int m_index;
		C *m_container;

		CMQIterator() : m_container(NULL), m_index(-1) {}
		CMQIterator(C *container, int index) : m_container(container), m_index(index) {
		}

		int getCount() const { return m_container->getCount<T>(); }
		bool isValid(int) const { return m_container->isValid<T>(index); }
		int next(int index, bool forward) const { return m_container->next<T>(index, forward); }
	};

	template <typename T, typename C>
	class CMQConstIterator : public std::iterator<std::bidirectional_iterator_tag, int>
	{
	public:
		friend class CMQObject;

		CMQConstIterator(const CMQConstIterator &iter) :
			m_container(iter.m_container), m_index(iter.m_index) {
		}

		CMQConstIterator &operator++() {
			m_index = next(m_index, true);
			return *this;
		}
		CMQConstIterator operator++(int) {
			CMQConstIterator origiter(*this);
			this->operator++();
			return origiter;
		}

		CMQConstIterator &operator--() {
			m_index = next(m_index, false);
			return *this;
		}
		CMQConstIterator operator--(int) {
			CMQConstIterator<T> origiter(*this);
			this->operator--();
			return origiter;
		}

		bool operator!=(const CMQConstIterator &iter) const { return m_index != iter.m_index; }
		bool operator==(const CMQConstIterator &iter) const { return !(this->operator!=(iter)); }

		//	const T &operator*() const {return get();}
		//	T &operator*() {return const_cast<T&>(get());}
		std::shared_ptr<const T> operator->() {
			std::shared_ptr<const T> ptr(getp());
			return ptr;
		}
		//	T *operator->() {return &get(m_index);}

		int index() const { return m_index; }
		//const T &get() const;
		std::shared_ptr<const T> getp() const;

	private:
		int m_index;
		const C *m_container;

		CMQConstIterator() : m_container(NULL), m_index(-1) {}
		CMQConstIterator(const C *container, int index) :
			m_container(container), m_index(index)
		{
		}

		int getCount() const { return m_container->getCount<T>(); }
		bool isValid(int) const { return m_container->isValid<T>(index); }
		int next(int index, bool forward) const { return m_container->next<T>(index, forward); }
	};

	class CMQDocument
	{
		friend class CMQObject;
		friend class Vertex;
		friend class Face;
	public:
		//	CMQDocument() : m_doc(nullptr) {}
		CMQDocument(const MQDocument doc) : m_doc(doc) {}
		CMQDocument(const CMQDocument &doc) : m_doc(doc.m_doc) {}
		CMQDocument &operator=(const CMQDocument &doc) { m_doc = doc.m_doc; return *this; }

		cmq::oindex_t getCurrentObjectIndex();
		CMQObject getCurrentObject();
		void setCurrentObjectIndex(cmq::oindex_t);
		void setCurrentObject(CMQObject &obj);
		CMQObject &addObject(CMQObject &newobj);
		CMQObject &insertObject(CMQObject &newobj, const CMQObject &before);
		void deleteObject(oindex_t i) { m_doc->DeleteObject(i); }
		void deleteObject(CMQObject &obj);
		void compact() { m_doc->Compact(); }

		// nameをベースとして使われていない名前を取得してstringで返す
		std::string getUnusedName(const std::string &name) const;

		int getObjectCount() const { return m_doc->GetObjectCount(); }

		CMQObject getObject(oindex_t i) const;
		CMQObject getObjectFromUniqueID(UINT uid) const;

		oindex_t getObjectIndex(CMQObject &obj) const;
		CMQObject getObjectbyName(const std::string &name) const;

		mindex_t getCurrentMaterialIndex() const { return m_doc->GetCurrentMaterialIndex(); }
		int getMaterialCount() const { return m_doc->GetMaterialCount(); }
		MQMaterial getMaterial(mindex_t i) const { return MQMaterial(m_doc->GetMaterial(i)); }
		void addMaterial(const MQMaterial &m) { m_doc->AddMaterial(m); }
		std::string getMaterialName(MQMaterial mat) const;
		void setMaterialName(MQMaterial mat, const std::string &name);
		std::string getTextureName(MQMaterial mat) const;

		bool isSelectMaterial(MQMaterial mat) const { return mat->GetSelected() == TRUE; }
		void setSelectMaterial(MQMaterial mat, bool flag) { mat->SetSelected(flag == TRUE); }

		bool isSelectVertex(const CMQObject &obj, vindex_t vindex) const;
		bool addSelectVertex(CMQObject &obj, vindex_t vindex);
		bool deleteSelectVertex(CMQObject &obj, vindex_t vindex);

		bool isSelectFace(const CMQObject &obj, findex_t findex) const;
		bool addSelectFace(CMQObject &obj, findex_t findex);
		bool deleteSelectFace(CMQObject &obj, findex_t findex);

		bool isSelectLine(const CMQObject &obj, findex_t findex, int lindex) const;
		bool addSelectLine(CMQObject &obj, findex_t findex, int lindex);
		bool deleteSelectLine(CMQObject &obj, findex_t findex, int lindex);

		bool isSelectUVVertex(const CMQObject &obj, findex_t findex, vindex_t vindex) const;
		bool addSelectUVVertex(CMQObject &obj, findex_t findex, vindex_t vindex);
		bool deleteSelectUVVertex(CMQObject &obj, findex_t findex, vindex_t vindex);

		void clearSelect(DWORD flag) { m_doc->ClearSelect(flag); }

		std::vector<TrianglePoints> triangulate(const std::vector<MQPoint> &parray);

		MQScene getScene(int index) const { return m_doc->GetScene(0); }

		template <typename T> int getCount() const;

		template <typename T> int next(int index, bool forward) {
			int count = getCount<T>();
			do {
				if(forward) index++; else index--;
			} while(index >= 0 && index < count && !isValid<T>(index));
			return index;
		}
/*
		template <typename T> CMQIterator<T> begin()
		{
			int index = 0;
			while(index < getCount<T>() && !isValid<T>(index)) {
				index = next<T>(index, true);
			}
			return CMQIterator<T>(this, index);
		}

		template <typename T> CMQIterator<T> end()
		{
			return const_cast<CMQObject*>(this)->end();
		}

		template <typename T> CMQConstIterator<T> cbegin() const
		{
			return const_cast<CMQObject*>(this)->begin();
		}

		template <typename T> CMQConstIterator<T> cend()
		{
			return const_cast<CMQObject*>(this)->end();
		}
*/
	private:
		MQDocument m_doc;
	};

	class CMQObject
	{
		friend class CMQDocument;
		friend class Vertex;
		friend class Face;
	public:
		CMQObject(CMQDocument *doc, oindex_t oindex) : m_obj(doc->m_doc->GetObject(oindex)), m_doc(*doc) {}
		CMQObject(MQObject obj, CMQDocument doc) : m_obj(obj), m_doc(doc) {}
		CMQObject(const CMQObject &obj) : m_obj(obj.m_obj), m_doc(obj.m_doc) {}
		CMQObject(const CMQObject &&obj) : m_obj(obj.m_obj), m_doc(obj.m_doc) {}
		CMQObject() : m_obj(nullptr), m_doc(nullptr) {}
		CMQObject &operator=(const CMQObject &obj) { m_obj = obj.m_obj; m_doc = obj.m_doc; return *this; }
		CMQObject &operator=(const CMQObject &&obj) { m_obj = obj.m_obj; m_doc = obj.m_doc; return *this; }
		bool operator==(const CMQObject &obj) const { return m_obj == obj.m_obj; }
		bool operator!=(const CMQObject &obj) const { return m_obj != obj.m_obj; }
		virtual ~CMQObject() {
			m_doc = nullptr;
			m_obj = nullptr;
		}
		void setDocument(CMQDocument doc) { m_doc = doc; }
		bool isLive() const;
		CMQObject clone();
		std::string getName() const;
		void setName(const std::string &name);
		UINT getUniqueID() const { return m_obj->GetUniqueID(); }
		bool getSelected() const { return !!m_obj->GetSelected(); }
		bool getLocking() const { return !!m_obj->GetLocking(); }
		bool getVisible() const { return !!m_obj->GetVisible(); }
		bool getFolding() const { return !!m_obj->GetFolding(); }
		void setSelected(bool flag) { m_obj->SetSelected(flag); }
		void setLocking(bool flag) { m_obj->SetLocking(flag); }
		void setVisible(bool flag) { m_obj->SetVisible(flag ? 0xFFFFFFFF : 0); }
		void setVisible(DWORD flag) { m_obj->SetVisible(flag); }
		void setFolding(bool flag) { m_obj->SetFolding(flag); }
		int getDepth() const { return m_obj->GetDepth(); }
		void setDepth(int depth) { m_obj->SetDepth(depth); }
		void deleteThis() { m_obj->DeleteThis(); }
		int getMirrorType() const { return m_obj->GetMirrorType(); }
		DWORD getMirrorAxis() const { return m_obj->GetMirrorAxis(); }
		float getMirrorDistance() const { return m_obj->GetMirrorDistance(); }
		int getLatheType() const { return m_obj->GetLatheType(); }
		DWORD getLatheAxis() const { return m_obj->GetLatheAxis(); }
		int getLatheSegment() const { return m_obj->GetLatheSegment(); }
		DWORD getPatchType() const { return m_obj->GetPatchType(); }
		int getPatchSegment() const { return m_obj->GetPatchSegment(); }
		BOOL getPatchTriangle() const { return m_obj->GetPatchTriangle(); }
		BOOL getShading() const { return m_obj->GetShading(); }
		float getSmoothAngle() const { return m_obj->GetSmoothAngle(); }

		void freeze(DWORD flag) { m_obj->Freeze(flag); }
		void compact() { m_obj->Compact(); }

		// 自分へオブジェクトコピー(頂点追加)する。指定されていればマトリックス変換を適用する
		void copyFromObject(const CMQObject &object, const MQMatrix *r = nullptr);

		int getVertexCount() const { return m_obj->GetVertexCount(); }
		int getVertexRefCount(vindex_t vindex) const { return m_obj->GetVertexRefCount(vindex); }
		MQPoint getVertex(vindex_t vindex) const { return m_obj->GetVertex(vindex); }
		UINT getVertexUniqueID(vindex_t vindex) const { return m_obj->GetVertexUniqueID(vindex); }
		vindex_t getVertexIndexFromUniqueID(UINT uid) const { return m_obj->GetVertexIndexFromUniqueID(uid); }
		float getVertexWeight(vindex_t vindex) const { return m_obj->GetVertexWeight(vindex); }
		void setVertex(vindex_t vindex, MQPoint p) { m_obj->SetVertex(vindex, p); }
		void setVertexWeight(vindex_t vindex, float weight) const { m_obj->SetVertexWeight(vindex, weight); }
		vindex_t addVertex(const MQPoint &p) { return m_obj->AddVertex(p); }
//		vindex_t addVertex(const CMQVertex &v) { return addVertex(v.p()); }
		BOOL deleteVertex(vindex_t index) { return m_obj->DeleteVertex(index); }

		bool isSelectVertex(vindex_t index) const;
		bool addSelectVertex(vindex_t index);
		bool deleteSelectVertex(vindex_t index);
		void copyVertexAttribute(vindex_t vert1, const CMQObject &obj2, vindex_t vert2) { m_obj->CopyVertexAttribute(vert1, obj2.m_obj, vert2); }

		int getFaceCount() const { return m_obj->GetFaceCount(); }
		int getFacePointCount(findex_t findex) const { return m_obj->GetFacePointCount(findex); }
		UINT getFaceUniqueID(findex_t findex) const { return m_obj->GetFaceUniqueID(findex); }
		findex_t getFaceIndexFromUniqueID(UINT uid) const { return m_obj->GetFaceIndexFromUniqueID(uid); }
		bool getFaceVisible(findex_t findex) const { return m_obj->GetFaceVisible(findex) == TRUE; }
		void setFaceVisible(findex_t findex, bool visible) { m_obj->SetFaceVisible(findex, visible ? TRUE : FALSE); }
		findex_t addFace(std::vector<vindex_t> varray);
		BOOL deleteFace(findex_t index) { return m_obj->DeleteFace(index); }
		int modifyFace(findex_t index, const std::vector<vindex_t> &varray);
		// 指定面の指定頂点を含む辺のリストを取得
		std::vector<findex_t> getVertexRelatedFaces(vindex_t vindex, std::vector<findex_t> &faces = std::vector<findex_t>(5)) const;
		// 面を構成する頂点のリストを取得
		std::vector<vindex_t> getFacePointArray(findex_t findex, std::vector<vindex_t> &vertexes = std::vector<vindex_t>(5)) const;
		mindex_t getFaceMaterial(findex_t findex) const { return m_obj->GetFaceMaterial(findex); }
		void setFaceMaterial(findex_t findex, mindex_t mindex) { m_obj->SetFaceMaterial(findex, mindex); }
		std::vector<CMQCoordinate> getFaceCoordinateArray(findex_t findex, std::vector<CMQCoordinate> &coord = std::vector<CMQCoordinate>(5)) const;
		void setFaceCoordinate(findex_t findex, vindex_t vertex, const CMQCoordinate &co);
		void setFaceCoordinateArray(findex_t findex, const std::vector<CMQCoordinate> &carray);
		DWORD getFaceVertexColor(findex_t findex, findex_t vindex) const { return m_obj->GetFaceVertexColor(findex, vindex); }
		void setFaceVertexColor(findex_t findex, findex_t vindex, DWORD color) { m_obj->SetFaceVertexColor(findex, vindex, color); }
		float getFaceEdgeCrease(findex_t findex, int lindex) const { return m_obj->GetFaceEdgeCrease(findex, lindex); }
		void setFaceEdgeCrease(findex_t findex, int lindex, float crease) { m_obj->SetFaceEdgeCrease(findex, lindex, crease); }
		bool isSelectFace(findex_t index) const;
		bool addSelectFace(findex_t index);
		bool deleteSelectFace(findex_t index);
		void copyFaceAttribute(findex_t findex, const CMQObject &object, findex_t findex2, std::vector<cmq::CMQCoordinate> &coord = std::vector<cmq::CMQCoordinate>(5));
		BOOL invertFace(findex_t index) { return m_obj->InvertFace(index); }

		bool isSelectLine(findex_t index, int lindex) const;
		bool addSelectLine(findex_t index, int lindex);
		bool deleteSelectLine(findex_t index, int lindex);

		bool isSelectUVVertex(findex_t findex, int index) const;
		bool addSelectUVVertex(findex_t findex, int index);
		bool deleteSelectUVVertex(findex_t findex, int index);

		bool isSelectUVVertex(flindex_t flindex) const { auto flpair = flId2FaceLinePair(flindex); return isSelectUVVertex(flpair.first, flpair.second); }
		bool addSelectUVVertex(flindex_t flindex) { auto flpair = flId2FaceLinePair(flindex); return addSelectUVVertex(flpair.first, flpair.second); }
		bool deleteSelectUVVertex(flindex_t flindex) { auto flpair = flId2FaceLinePair(flindex); return deleteSelectUVVertex(flpair.first, flpair.second); }

		// 面インデックスと辺インデックスで指定したラインの両端点の頂点インデックスを取得
		std::pair<vindex_t, vindex_t> getLineVertex(findex_t findex, int lindex) const;
	
		// 面+辺インデックスで指定したラインの両端点の頂点インデックスを取得
		std::pair<vindex_t, vindex_t> getLineVertex(flindex_t flindex) const { auto pair = lineId2VertexPair(flindex); return getLineVertex(pair.first, pair.second); }
		
		// 面インデックスと辺インデックスで指定したラインをv0を始点としたベクトルで取得
		MQPoint getLineVector(findex_t findex, int lindex, vindex_t v0 = -1) const;

		// 面+辺インデックスで指定したラインをv0を始点としたベクトルで取得
		MQPoint getLineVector(flindex_t flindex, vindex_t v0 = -1) const { auto pair = lineId2VertexPair(flindex); return getLineVector(pair.first, pair.second, v0); }
		
		// 面の指定した辺における法線ベクトルを取得
		MQPoint getNormalVector(findex_t findex, int lindex = 0) const;

		// 面の指定辺の対角頂点の座標
		MQPoint getNeiborPoint(findex_t findex, int lindex) const;

		MQPoint getVector(vindex_t v1, vindex_t v2, vindex_t v0) const { return (v0 == v2) ? getVertex(v1) - getVertex(v2) : getVertex(v2) - getVertex(v1); }
		MQPoint getVector(lindex_t line, vindex_t v0 = -1) const { auto pair = lineId2VertexPair(line); return getVector(pair.first, pair.second, v0); }
		std::unordered_set<flindex_t> getNeighborLine(vindex_t vindex, lindex_t lindex = -1) const;
		std::unordered_set<vindex_t> getNeighborVertex(vindex_t vindex) const;
		std::unordered_set<flindex_t> getLineWithVertex(vindex_t vindex1, vindex_t vindex2, flindex_t flindex_except = -1) const;
		findex_t createLine(vindex_t v1, vindex_t v2);
		findex_t createLine(vindex_t v1, MQPoint &p2);
		findex_t createLine(vindex_t v1, MQPoint &p2, vindex_t &added_vertex);
		findex_t createLine(MQPoint &p1, MQPoint &p2);
		findex_t createLine(MQPoint &p1, MQPoint &p2, vindex_t added_vertex[2]);
		std::vector<std::pair<findex_t, int>> getOtherLine(vindex_t v1, vindex_t v2, findex_t face = -1, bool selectonly = false);
		std::vector<std::pair<findex_t, int>> getOtherLine(std::pair<vindex_t, vindex_t> line, findex_t face = -1, bool selectonly = false);
		std::vector<lindex_t> getLineArray(findex_t findex, vindex_t vindex = -1);

		MQPoint getTranslation() const { return m_obj->GetTranslation(); }
		void setTranslation(const MQPoint &p) { m_obj->SetTranslation(p); }
		MQAngle getRotation() const { return m_obj->GetRotation(); }
		void setRotation(const MQAngle &a) { m_obj->SetRotation(a); }
		MQMatrix getLocalMatrix() const { return m_obj->GetLocalMatrix(); }
		void setLocalMatrix(const MQMatrix &r) { m_obj->SetLocalMatrix(r); }
#if MQPLUGIN_VERSION >= 0x0400
		std::vector<TrianglePoints> triangulate(findex_t findex);
#endif
		template <typename T> int getCount() const;
		template <typename T> bool isValid(int index) const;
		//	template <class T> T get(int index) const;

		template <typename T> int next(int index, bool forward) const {
			int count = getCount<T>();
			do {
				if(forward) index++; else index--;
			} while(index >= 0 && index < count && !isValid<T>(index));
			return index;
		}

		template <typename T, typename P> CMQIterator<T, P> begin() {
			int index = 0;
			while(index < getCount<T>() && !isValid<T>(index)) {
				index = next<T>(index, true);
			}
			return CMQIterator<T, P>(this, index);
		}

		template <typename T, typename P> CMQConstIterator<T, P> cbegin() const {
			int index = 0;
			while (index < getCount<T>() && !isValid<T>(index)) {
				index = next<T>(index, true);
			}
			return CMQConstIterator<T, P>(this, index);
		}

		template <typename T, typename P> CMQIterator<T, P> end() {
			CMQIterator<T, P> iter(this, getCount<T>());
			return iter;
		}

		template <typename T, typename P> CMQConstIterator<T, P> cend() const {
			CMQConstIterator<T, P> iter(this, getCount<T>());
			return iter;
		}

		MQMatrix getLocal() const { return m_obj->GetLocalMatrix(); }
		MQMatrix getLocalInverseMatrix() const { return m_obj->GetLocalInverseMatrix(); }
		void setLocal(const MQMatrix &r) { m_obj->SetLocalMatrix(r); }

		//private:
		MQObject m_obj;
		CMQDocument m_doc;
	};

	template <typename T, typename P>
	class CMQContainer
	{
	public:
		CMQContainer(P *pobj) : m_pobj(pobj) {}

		CMQIterator<T, P> begin() { return m_pobj->begin<T, P>(); }
		CMQIterator<T, P> end() { return m_pobj->end<T, P>(); }

	private:
		P *m_pobj;
	};

	template <typename T, typename P>
	class CMQConstContainer
	{
	public:
		CMQConstContainer(const P *pobj) : m_pobj(pobj) {}

		CMQConstIterator<T, P> begin() { return m_pobj->cbegin<T, P>(); }
		CMQConstIterator<T, P> end() { return m_pobj->cend<T, P>(); }

	private:
		const P *m_pobj;
	};
/*
	template <class T, class P>
	class CMQConstContainer
	{
	public:
		CMQConstContainer(const P *pobj) : m_pobj(pobj) {}

		const CMQIterator<T> cbegin() const { return m_pobj->cbegin<T>(); }
		const CMQIterator<T> cend() const { return m_pobj->cend<T>(); }

	private:
		const P *m_pobj;
	};
*/
/*
	template <class T, class P>
	class CMQContainer
	{
	public:
		CMQContainer(const P *pobj) : m_pobj(pobj) {}

		const CMQIterator<T> cbegin() const { return m_pobj->cbegin<T>(); }
		const CMQIterator<T> cend() const { return m_pobj->cend<T>(); }

	private:
		const P *m_pobj;
	};
*/
	std::wstring towstring(const std::string &s);

	std::string tostring(const std::wstring &s);

}

namespace std
{
	template <>
	struct hash<cmq::CMQLindex>
	{
		std::size_t operator () (cmq::CMQLindex const & key) const
		{
			return key.hash();
		}
	};
}

#endif
